from reportlab.lib.units import mm
from reportlab.lib.pagesizes import A4
from reportlab.pdfgen import canvas


class AlgorithmPdfGenerator(object):
    """
    Class for creating simple reports of processing of dynamic programming algorithms.

    Uses ReportLab library and many magical constants to render PDF correctly.
    """
    def __init__(self, filename):
        """
        Creates new generator that saves its output to file filename.
        """
        self._page_size = A4
        self.c = canvas.Canvas(filename, pagesize=self._page_size)
        self.box_size = 10  # size of square with single DP element
        self.box_spacing = 2  # spacing between elements

    @property
    def grid_size(self):
        return self.box_spacing + self.box_size

    def _setLineWidth(self, bold):
        """
        Switches between thick and thin line.
        """
        if bold:
            self.c.setLineWidth(3)
        else:
            self.c.setLineWidth(1)

    def _draw_matrix(self, algorithm, alignment):
        """
        Create matrix section.
        """
        self.c.setFont('Helvetica', 20)
        self.c.setStrokeColorRGB(0, 0, 0)
        self.c.setFillColorRGB(0, 0, 0)

        matrix = algorithm.dp_matrix
        if not matrix or matrix.size_m <= 1 or matrix.size_n <= 1:
            return

        # draw vertical letters
        for i, char in enumerate(algorithm.sequence1):
            self.c.drawCentredString((self.box_size / 2) * mm,
                                     - ((i + 2) * self.grid_size + self.box_size / 1.3) * mm,
                                     unicode(char))

        # draw horizontal letters
        for i, char in enumerate(algorithm.sequence2):
            self.c.drawCentredString(((i + 2) * self.grid_size + self.box_size / 2.3) * mm,
                                     - self.box_size * mm,
                                     unicode(char))

        self.c.setFontSize(12)

        # draw matrix
        for j in range(0, matrix.size_n):
            for i in range(0, matrix.size_m):
                # painter.setPen(QtGui.QPen(QtGui.QColor(20, 20, 20), 1, QtCore.Qt.SolidLine))
                element = matrix[i][j]

                # highlight this element if it's part of the alignment track
                self.c.setFillColorRGB(0.8, 0.8, 0.8)
                fill = 1 if alignment and (i, j) in alignment.track else 0

                self.c.rect((j + 1) * self.grid_size * mm,
                            - (i + 1) * self.grid_size * mm,
                            self.box_size * mm,
                            - self.box_size * mm,
                            fill=fill)

                self.c.setFillColorRGB(0, 0, 0)
                self.c.drawCentredString(((j + 1) * self.grid_size + self.box_size / 2) * mm,
                                         - ((i + 1) * self.grid_size + self.box_size / 1.6) * mm,
                                         unicode(element.value))

                track = alignment.track
                if element.from_diagonal:
                    self._setLineWidth((i, j) in track and (i - 1, j - 1) in track)
                    self.c.line((j + 1) * self.grid_size * mm,
                                - (i + 1) * self.grid_size * mm,
                                ((j + 1) * self.grid_size - self.box_spacing) * mm,
                                - ((i + 1) * self.grid_size - self.box_spacing) * mm)
                if element.from_horizontal:
                    self._setLineWidth((i, j) in track and (i - 1, j) in track)
                    self.c.line(((j + 1) * self.grid_size + self.box_size / 2) * mm,
                                - (i + 1) * self.grid_size * mm,
                                ((j + 1) * self.grid_size + self.box_size / 2) * mm,
                                - ((i + 1) * self.grid_size - self.box_spacing) * mm)
                if element.from_vertical:
                    self._setLineWidth((i, j) in track and (i, j - 1) in track)
                    self.c.line(((j + 1) * self.grid_size) * mm,
                                - ((i + 1) * self.grid_size + self.box_size / 2) * mm,
                                ((j + 1) * self.grid_size - self.box_spacing) * mm,
                                - ((i + 1) * self.grid_size + self.box_size / 2) * mm)
                self._setLineWidth(False)

    def _draw_header(self):
        """
        Create header.
        """
        self.c.setFont('Helvetica', 32)
        self.c.drawCentredString(0, 20 * mm, 'SeAlignator report')

    def _draw_parameters(self, algorithm):
        """
        Create section with algorithm parameters.
        """
        self.c.setFontSize(12)
        parameters_text = "Algorith: %s\n" % algorithm.name
        if hasattr(algorithm, 'substitution_matrix'):
            parameters_text += "Substitution matrix: %s\n" % algorithm.substitution_matrix.name
        for parameter, properties in algorithm.parameters.iteritems():
            if properties['type'] == bool:
                continue
            value = getattr(algorithm, parameter)
            parameters_text += "%s: %s\n" % (properties['name'], value)

        for i, line in enumerate(parameters_text.split('\n')):
            self.c.drawString(0, - i * 12, line)

    def _draw_alignment(self, alignment):
        """
        Create alignment section of document.
        """
        self.c.setFont('Helvetica', 12)
        self.c.drawString(0, 0, 'Alignment:')
        self.c.setFont('Courier', 16)
        self.c.drawString(0, - 15, alignment.sequence1)
        self.c.drawString(0, - 30, alignment.sequence2)

    def create_report(self, algorithm, alignment):
        """
        Create report with algorithm parameters, matrix and visualization
        of supplied alignment.
        """
        self.c.translate(self._page_size[0] / 2 + 5 * mm, self._page_size[1] - 40 * mm)
        self._draw_header()
        self.c.translate(- self._page_size[0] / 2 + 5 * mm, 12 * mm)
        self._draw_parameters(algorithm)
        self.c.translate(0, - 25 * mm)
        self._draw_alignment(alignment)
        self.c.translate(- 8 * mm, - 10 * mm)
        self._draw_matrix(algorithm, alignment)
        self.c.showPage()
        self.c.save()