import inspect
import os
import string
import sys
from PySide import QtGui, QtCore
from PySide.QtCore import Qt
from bioinformatics.algorithms import *
from bioinformatics.matrices import MatrixParsingException, SubstitutionMatrix
from pdf_generator import AlgorithmPdfGenerator

version = "1.0.1"


class SequenceValidator(QtGui.QValidator):
    """
    QValidator that validates that sequence consists only of allowed characters.
    """
    def __init__(self, parent=None):
        super(SequenceValidator, self).__init__(parent)
        self.allowedChars = None  # None stands for every char, empty string for no chars

    def fixup(self, s):
        """
        Returns new string containing only allowed characters.
        """
        if self.allowedChars is None:
            return s

        table = string.maketrans(''.join(self.allowedChars), ' ' * len(self.allowedChars))
        return str(s.encode('utf-8')).translate(None, table)

    def validate(self, s, pos):
        if self.allowedChars is None:
            return QtGui.QValidator.Acceptable, s, pos

        s = s.upper()
        if pos == 0 or s[pos - 1] in self.allowedChars:
            return QtGui.QValidator.Acceptable, s, pos
        else:
            return QtGui.QValidator.Invalid, s, pos


class DPMatrixWidget(QtGui.QWidget):
    """
    Widget that visualizes matrices of dynamic programing algorithms.
    """
    alignmentChanged = QtCore.Signal()

    def __init__(self, parent):
        super(DPMatrixWidget, self).__init__(parent)
        self.algorithm = None
        self.alignment = None
        self.boxSize = 30  # size of square with single DP element
        self.boxSpacing = 10  # spacing between elements
        self.setMouseTracking(True)  # mouse tracking must be enabled for tooltips
        self.selectedElement = (-1, -1)  # element to backtrack from

    @property
    def gridSize(self):
        return self.boxSize + self.boxSpacing

    def updateAlignment(self, alignmentIndex):
        """
        Highlight elements of current alignment.
        """
        if len(self.algorithm.alignments) < 1:
            return
        self.alignment = self.algorithm.alignments[alignmentIndex]
        self.update()

    def sizeHint(self):
        if self.algorithm and self.algorithm.dp_matrix:
            h = (self.algorithm.dp_matrix.size_m + 1) * self.gridSize - self.boxSpacing // 2
            w = (self.algorithm.dp_matrix.size_n + 1) * self.gridSize - self.boxSpacing // 2
            return QtCore.QSize(w, h)

        return QtCore.QSize(0, 0)

    def _get_element_position_at(self, x, y):
        """
        Returns (i, j) tuple representing element index on position
        x, y of widget.
        """
        if y < self.gridSize or x < self.gridSize:
            # one gridSize is reserved for letters, do nothing
            return None

        i = (y - self.gridSize) / self.gridSize
        j = (x - self.gridSize) / self.gridSize
        matrix = self.algorithm.dp_matrix
        if matrix and matrix.size_m > i and matrix.size_n > j:
            return i, j

        return None

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            position = self._get_element_position_at(event.x(), event.y())

            if position is not None:
                self.selectedElement = position
                i, j = position
                self.algorithm.arbitrary_backtrack(i, j)
                self.update()
                self.alignmentChanged.emit()
        else:
            super(DPMatrixWidget, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        widgetPosition = self.mapFromGlobal(event.globalPos())
        position = self._get_element_position_at(widgetPosition.x(), widgetPosition.y())

        if position is not None:
            i, j = position
            text = "<p>"
            if i == 0 and j == 0:
                text += "First element equals always zero."
            elif i == 0:
                text += "i = %s, j = %s<br>%s" % (i, j, self.algorithm._get_initial_i.__doc__)
            elif j == 0:
                text += "i = %s, j = %s<br>%s" % (i, j, self.algorithm._get_initial_j.__doc__)
            else:
                text += "i = %s, j = %s<br>" % (i, j)
                text += '%s<br><br>Match/mismatch score = <b>%s</b><br>'\
                        % (self.algorithm._get_dp_matrix_value_at.__doc__,
                           self.algorithm._calculate_diagonal_score(i, j))
                if 'affine_gap_cost' in self.algorithm.parameters:
                    h_doc = self.algorithm._get_h_matrix_value_at.__doc__
                    v_doc = self.algorithm._get_v_matrix_value_at.__doc__
                    text += '<br>H-matrix<br>%s<br>H(i, j) = <b>%s</b><br><br>V-matrix<br>%s<br>V(i, j) = <b>%s</b>'\
                            % (h_doc, self.algorithm.h_matrix[i][j], v_doc, self.algorithm.v_matrix[i][j])
            QtGui.QToolTip.showText(event.globalPos(), text + "</p>", self)

    def _setPainterPenStyle(self, painter, dark):
        """
        Sets painter's pen to dark or light color.
        """
        if dark:
            painter.setPen(QtGui.QPen(QtGui.QColor(20, 20, 20), 2, QtCore.Qt.SolidLine))
        else:
            painter.setPen(QtGui.QPen(QtGui.QColor(200, 200, 200), 2, QtCore.Qt.SolidLine))

    def paintEvent(self, event):
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setBrush(QtCore.Qt.NoBrush)

        matrix = self.algorithm.dp_matrix
        if not matrix or matrix.size_m <= 1 or matrix.size_n <= 1:
            return

        font = painter.font()
        old_size = font.pointSize()
        font.setPointSize(16)
        painter.setFont(font)

        # draw vertical letters
        for i, char in enumerate(self.algorithm.sequence1):
            painter.drawText(0, (i + 2) * self.gridSize,
                             self.boxSize, self.boxSize,
                             Qt.AlignCenter, unicode(char))

        # draw horizontal letters
        for i, char in enumerate(self.algorithm.sequence2):
            painter.drawText((i + 2) * self.gridSize, 0,
                             self.boxSize, self.boxSize,
                             Qt.AlignCenter, unicode(char))

        # reset font to default
        font.setPointSize(old_size)
        painter.setFont(font)

        # draw matrix
        for j in range(0, matrix.size_n):
            for i in range(0, matrix.size_m):
                painter.setPen(QtGui.QPen(QtGui.QColor(20, 20, 20), 1, QtCore.Qt.SolidLine))
                element = matrix[i][j]
                painter.drawRect((j + 1) * self.gridSize, (i + 1) * self.gridSize,
                                 self.boxSize, self.boxSize)

                # highlight this element if it's part of the alignment track
                if self.alignment and (i, j) in self.alignment.track:
                    painter.fillRect((j + 1) * self.gridSize + 1, (i + 1) * self.gridSize + 1,
                                     self.boxSize - 1, self.boxSize - 1, QtCore.Qt.lightGray)

                painter.drawText((j + 1) * self.gridSize, (i + 1) * self.gridSize,
                                 self.boxSize, self.boxSize,
                                 Qt.AlignCenter, unicode(element.value))

                # paint lines from this element to previous if we have an alignment
                if not self.alignment:
                    return
                track = self.alignment.track
                if element.from_diagonal:
                    self._setPainterPenStyle(painter, (i, j) in track and (i - 1, j - 1) in track)
                    painter.drawLine((j + 1) * self.gridSize, (i + 1) * self.gridSize,
                                     (j + 1) * self.gridSize - self.boxSpacing,
                                     (i + 1) * self.gridSize - self.boxSpacing)
                if element.from_horizontal:
                    self._setPainterPenStyle(painter, (i, j) in track and (i - 1, j) in track)
                    painter.drawLine((j + 1) * self.gridSize + self.boxSize // 2, (i + 1) * self.gridSize,
                                     (j + 1) * self.gridSize + self.boxSize // 2,
                                     (i + 1) * self.gridSize - self.boxSpacing + 1)
                if element.from_vertical:
                    self._setPainterPenStyle(painter, (i, j) in track and (i, j - 1) in track)
                    painter.drawLine((j + 1) * self.gridSize, (i + 1) * self.gridSize + self.boxSize // 2,
                                     (j + 1) * self.gridSize - self.boxSpacing + 1,
                                     (i + 1) * self.gridSize + self.boxSize // 2)

    def runAlgorithm(self, sequence1, sequence2, **kwargs):
        """
        Run current algorithm and update matrix.
        """
        self.algorithm.run(sequence1, sequence2, **kwargs)
        self.alignmentChanged.emit()
        self.adjustSize()
        self.update()


class SeAlignator(QtGui.QMainWindow):
    def __init__(self):
        super(SeAlignator, self).__init__()
        self.setMinimumSize(800, 600)
        centralWidget = QtGui.QWidget()

        # sequence editation
        sequence1Label = QtGui.QLabel("Sequence 1")
        self.sequence1Edit = QtGui.QLineEdit()
        sequence2Label = QtGui.QLabel("Sequence 2")
        self.sequence2Edit = QtGui.QLineEdit()
        # set custom validator
        self.sequenceValidator = SequenceValidator()
        self.sequence1Edit.setValidator(self.sequenceValidator)
        self.sequence2Edit.setValidator(self.sequenceValidator)

        # alignment selection
        alignmentLabel = QtGui.QLabel("Alignments")
        self.alignmentCombo = QtGui.QComboBox()
        self.alignmentExportButton = QtGui.QPushButton("Export alignment")
        self.alignmentExportButton.setEnabled(False)

        self.matrixLabel = QtGui.QLabel("Matrix: ?")
        self.importMatrixButton = QtGui.QPushButton("Import matrix")

        # last selected file containing matrix data
        self.matrixFile = None
        self.substitutionMatrix = None
        self._loadSubstitutionMatrix('data/matrices/PAM250')

        # algorithm selection
        self.algorithms = []
        algorithmLabel = QtGui.QLabel("Algorithm:")
        self.algorithmCombo = QtGui.QComboBox()
        self.findAlgorithms()

        # matrix display
        self.scrollArea = QtGui.QScrollArea()
        self.matrixWidget = DPMatrixWidget(self)
        self.changeAlgorithm(0, False)
        self.createParametersGroup()

        self.scrollArea.setWidget(self.matrixWidget)

        self.matrixWidget.alignmentChanged.connect(self.findAlignments)
        self.alignmentCombo.currentIndexChanged[int].connect(self.matrixWidget.updateAlignment)
        self.algorithmCombo.currentIndexChanged[int].connect(self.changeAlgorithm)
        self.sequence1Edit.textChanged.connect(self.recalculate)
        self.sequence2Edit.textChanged.connect(self.recalculate)
        self.importMatrixButton.clicked.connect(self.openMatrixButtonClicked)
        self.alignmentExportButton.clicked.connect(self.alignmentExportButtonClicked)

        self.controlsLayout = QtGui.QGridLayout()
        self.controlsLayout.addWidget(algorithmLabel, 0, 0)
        self.controlsLayout.addWidget(self.algorithmCombo, 0, 1)
        self.controlsLayout.addWidget(sequence1Label, 1, 0)
        self.controlsLayout.addWidget(self.sequence1Edit, 1, 1)
        self.controlsLayout.addWidget(sequence2Label, 2, 0)
        self.controlsLayout.addWidget(self.sequence2Edit, 2, 1)
        self.controlsLayout.addWidget(self.importMatrixButton, 3, 0)
        self.controlsLayout.addWidget(self.matrixLabel, 3, 1)
        self.controlsLayout.addWidget(self.parametersGroup, 4, 0, 1, 2)
        self.controlsLayout.addWidget(alignmentLabel, 5, 0)
        self.controlsLayout.addWidget(self.alignmentCombo, 5, 1)

        centralLayout = QtGui.QVBoxLayout()
        centralLayout.addLayout(self.controlsLayout)

        centralLayout.addWidget(self.scrollArea, 1)
        centralWidget.setLayout(centralLayout)
        centralLayout.addWidget(self.alignmentExportButton)

        self.setCentralWidget(centralWidget)
        self.setWindowTitle("SeAlignator %s" % version)

    def createParametersGroup(self):
        """
        Creates QGroupBox of current algorithm's parameters.
        """
        self.parametersGroup = QtGui.QGroupBox("Algorithm parameters")

        self.algorithmParameterLabels = {}
        self.algorithmParameters = {}

        parametersLayout = QtGui.QGridLayout()

        for parameter, properties in sorted(self.matrixWidget.algorithm.parameters.iteritems(),
                                            key=lambda x: x[1]['pos']):
            if properties['type'] == int:
                self.algorithmParameterLabels[parameter] = QtGui.QLabel(properties['name'])
                self.algorithmParameters[parameter] = QtGui.QSpinBox()
                self.algorithmParameters[parameter].setRange(properties['min'], properties['max'])
                self.algorithmParameters[parameter].setSingleStep(1)
                self.algorithmParameters[parameter].setValue(properties['default'])
                self.algorithmParameters[parameter].valueChanged.connect(self.recalculate)
                parametersLayout.addWidget(self.algorithmParameterLabels[parameter], properties['pos'], 0)
                parametersLayout.addWidget(self.algorithmParameters[parameter], properties['pos'], 1)
            elif properties['type'] == bool:
                self.algorithmParameters[parameter] = QtGui.QCheckBox(properties['name'])
                self.algorithmParameters[parameter].setChecked(properties['default'])
                parametersLayout.addWidget(self.algorithmParameters[parameter], properties['pos'], 0, 1, 2)
                p = parameter
                self.algorithmParameters[parameter].stateChanged.connect(
                    lambda: self._parameterToggleChanged(p))

        # hide matrix button if this algorithm does not uses substitution matrix
        if hasattr(self.matrixWidget.algorithm, 'substitution_matrix'):
            self.matrixLabel.show()
            self.importMatrixButton.show()
        else:
            self.matrixLabel.hide()
            self.importMatrixButton.hide()

        # stretch content closer to labels
        parametersLayout.setColumnStretch(1, 1)
        self.parametersGroup.setLayout(parametersLayout)

    def _parameterToggleChanged(self, parameter):
        """
        Method used for toggling 'toggles' of boolean parameters.
        """
        toggles = self.matrixWidget.algorithm.parameters[parameter]['toggles']
        self.algorithmParameters[toggles].setEnabled(self.algorithmParameters[parameter].isChecked())
        self.recalculate()

    def getAlgorithmParameters(self):
        """
        Returns dict of parameters set in GUI.
        """
        kwargs = {}
        for name, widget in self.algorithmParameters.iteritems():
            if isinstance(widget, QtGui.QCheckBox):
                pass  # boolean parameters can't be passed to the algorithm
            else:
                kwargs[name] = widget.value() if widget.isEnabled() else None
        return kwargs

    def findAlgorithms(self):
        """
        Find available algorithms in bioinformatics.algorithms module.
        """
        self.algorithms = []
        self.algorithmCombo.blockSignals(True)
        for name, value in inspect.getmembers(sys.modules[__name__], inspect.isclass):
            if issubclass(value, DPAlgorithm) and value is not DPAlgorithm:
                self.algorithms.append(value)
                self.algorithmCombo.addItem(value.name)
        self.algorithmCombo.blockSignals(False)

    def findAlignments(self):
        """
        Fetch all alignments and feed them to combo.
        """
        if self.matrixWidget.algorithm is None:
            return

        self.alignmentCombo.clear()
        for alignment in self.matrixWidget.algorithm.alignments:
            self.alignmentCombo.addItem(unicode(alignment))

    def _loadSubstitutionMatrix(self, filename):
        """
        Load substitution matrix from filename.
        """
        if self.substitutionMatrix is None:
            self.substitutionMatrix = SubstitutionMatrix()
        filename = os.path.abspath(filename)
        display_name = os.path.split(filename)[1]
        try:
            self.matrixFile = filename
            self.substitutionMatrix.load_matrix(filename)
            self.statusBar().showMessage("Loaded matrix file: %s" % display_name)
            self.matrixLabel.setText("Matrix: %s" % display_name)
            # update allowed chars in validator according to matrix
            self.sequenceValidator.allowedChars = self.substitutionMatrix._indexes.keys()
            self._fixupSequences()
        except (MatrixParsingException, IOError):
            self.statusBar().showMessage("Unable to load matrix %s" % display_name)
            QtGui.QMessageBox.warning(self, "Import error", "Unable to load matrix file %s." % filename)

    def _fixupSequences(self):
        """
        Runs sequence1 and sequence2 through the current validator's fixup method.
        """
        self.sequence1Edit.blockSignals(True)
        self.sequence2Edit.blockSignals(True)
        self.sequence1Edit.setText(self.sequenceValidator.fixup(self.sequence1Edit.text()))
        self.sequence2Edit.setText(self.sequenceValidator.fixup(self.sequence2Edit.text()))
        self.sequence1Edit.blockSignals(False)
        self.sequence2Edit.blockSignals(False)

    def changeAlgorithm(self, algorithmIndex, recalculate=True):
        """
        Change current algorithm to algorithm of index algorithmIndex in self.algorithms
        and recalculate (if recalculate parameter is not False).
        """
        algorithm = self.algorithms[algorithmIndex]
        self.matrixWidget.algorithm = algorithm()
        if hasattr(self.matrixWidget.algorithm, 'substitution_matrix'):
            self.matrixWidget.algorithm.substitution_matrix = self.substitutionMatrix
            self.sequenceValidator.allowedChars = self.substitutionMatrix._indexes.keys()
        else:
            self.sequenceValidator.allowedChars = None
        self._fixupSequences()
        if hasattr(self, 'parametersGroup'):
            self.parametersGroup.hide()
        self.createParametersGroup()
        if hasattr(self, 'controlsLayout'):
            self.controlsLayout.addWidget(self.parametersGroup, 4, 0, 1, 2)
        if recalculate:
            self.recalculate()
        self.update()

    def recalculate(self):
        kwargs = self.getAlgorithmParameters()
        self.matrixWidget.runAlgorithm(self.sequence1Edit.text(), self.sequence2Edit.text(), **kwargs)
        if len(self.matrixWidget.algorithm.alignments) > 0:
            self.alignmentExportButton.setEnabled(True)
        else:
            self.alignmentExportButton.setEnabled(False)
        self.update()

    def openMatrixButtonClicked(self):
        fileDialog = QtGui.QFileDialog(parent=self, directory='data/matrices')
        fileDialog.setFileMode(QtGui.QFileDialog.ExistingFile)
        if fileDialog.exec_():
            matrixFile = fileDialog.selectedFiles()[0]
            self._loadSubstitutionMatrix(matrixFile)
            self.recalculate()

    def alignmentExportButtonClicked(self):
        fileDialog = QtGui.QFileDialog(parent=self, directory='data')
        fileDialog.setNameFilter("PDF files (*.pdf)")
        fileDialog.setFileMode(QtGui.QFileDialog.AnyFile)
        fileDialog.setAcceptMode(QtGui.QFileDialog.AcceptSave)
        if fileDialog.exec_():
            path, ext = os.path.splitext(os.path.abspath(fileDialog.selectedFiles()[0]))
            if ext != 'pdf':
                ext = '.pdf'
            filename = path + ext
            try:
                g = AlgorithmPdfGenerator(filename)
                g.create_report(self.matrixWidget.algorithm, self.matrixWidget.alignment)
                self.statusBar().showMessage("Report saved to %s" % filename)
            except IOError:
                QtGui.QMessageBox.warning(self, "Export error", "Unable to write PDF report to %s" % filename)


if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)
    window = SeAlignator()
    window.show()
    sys.exit(app.exec_())