from matrices import DPMatrix


class PairwiseAlignment(object):
    """
    Alignment of two sequences and its track through DP matrix.
    """
    def __init__(self, sequence1="", sequence2="", track=None):
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.track = track or []

    def prepend_to_new(self, to_sequence1, to_sequence2, matrix_field_pos):
        """
        Prepends to_sequence1 to 1st and to_sequence2 to 2nd alignment (backtracking
        is done backwards) and returns new instance of PairwiseAlignment. Also updates
        new alignment's track by adding matrix_field_pos.
        """
        new_track = list(self.track)
        new_track.append(matrix_field_pos)
        return PairwiseAlignment(to_sequence1 + self.sequence1,
                                 to_sequence2 + self.sequence2, new_track)

    def __unicode__(self):
        return "%s | %s" % (self.sequence1, self.sequence2)

    def __str__(self):
        return unicode(self).encode('utf-8')


class Log(object):
    """
    Helper class used for logging of algorithm progress.
    """
    def __init__(self):
        self.sections = []

    def add_section(self, name):
        """
        Create new section to log into.
        """
        self.sections.append({'name': name, 'content': []})

    def append(self, item):
        """
        Add new log to current section.
        """
        if len(self.sections):
            self.add_section('Header')
        self.sections[-1]['content'].append(item.rstrip())

    def __unicode__(self):
        return '\n'.join("%s\n%s\n%s" % (section['name'], "=" * 80,
                                         '\n'.join(section['content'])) for section in self.sections)

    def __str__(self):
        return unicode(self).encode('utf-8')


class DPAlgorithm(object):
    """
    Base of dynamic programming algorithms.
    """
    max_alignments = 500

    def __init__(self):
        self.sequence1 = ""
        self.sequence2 = ""
        self.dp_matrix = None
        self.alignments = []
        self.log = Log()

    def _is_backtracking_done(self, i, j):
        """
        Backtracking is done when top left element is reached.
        """
        return i == 0 and j == 0

    def _initialize(self):
        """
        Matrices initialization - fills G with zeroes and calculates
        values for zeroth column and row.
        """
        self.log.add_section('Initialization')
        self.dp_matrix = DPMatrix(len(self.sequence1) + 1, len(self.sequence2) + 1)

        for i in range(1, self.dp_matrix.size_m):
            value = self._get_initial_i(i)
            self.log.append('Set G(%s, 0) = %s' % (i, value))
            self.dp_matrix[i][0].value = value
            self.dp_matrix[i][0].from_horizontal = True

        for j in range(1, self.dp_matrix.size_n):
            value = self._get_initial_j(j)
            self.log.append('Set G(0, %s) = %s' % (j, value))
            self.dp_matrix[0][j].value = value
            self.dp_matrix[0][j].from_vertical = True

    def _get_initial_i(self, i):
        """
        Equals i.
        """
        return i

    def _get_initial_j(self, j):
        """
        Equals j.
        """
        return j

    def _calculate_substitution_score(self, c1, c2):
        """
        Abstract (not implemented). Calculate substitution cost of c1 and c2.
        """
        raise NotImplementedError

    def _calculate_diagonal_score(self, i, j):
        """
        Calculates value of (mis)match as sum of diagonal (upper left) element
        and (mis)match scoring function.
        """
        return self.dp_matrix[i - 1][j - 1].value + self._calculate_substitution_score(self.sequence1[i - 1],
                                                                                self.sequence2[j - 1])

    def backtrack(self, i, j, alignment=None):
        """
        Run recursive backtracking from specified position.

        Returns up to DPAlgorithm.max_aligments alignments.
        """
        if not alignment:
            alignment = PairwiseAlignment("", "", [(i, j)])

        if len(self.alignments) >= DPAlgorithm.max_alignments:
            return

        if self._is_backtracking_done(i, j):
            self.alignments.append(alignment)
            return

        element = self.dp_matrix[i][j]

        if element.from_horizontal:
            self.backtrack(i - 1, j,
                           alignment.prepend_to_new(self.sequence1[i - 1], '-', (i - 1, j)))

        if element.from_vertical:
            self.backtrack(i, j - 1,
                           alignment.prepend_to_new('-', self.sequence2[j - 1], (i, j - 1)))

        if element.from_diagonal:
            self.backtrack(i - 1, j - 1,
                           alignment.prepend_to_new(self.sequence1[i - 1], self.sequence2[j - 1], (i - 1, j - 1)))

    def arbitrary_backtrack(self, i, j):
        """
        Clear alignments and start backtracking from specified position.
        """
        self.alignments = []
        self.backtrack(i, j)


class OWED(DPAlgorithm):
    """
    Implementation of operation-weighted edit distance.

    Based on David Hoksza's bioinformatics lectures: http://siret.ms.mff.cuni.cz/members/hoksza/lectures

    Notation - G matrix (dp_matrix) - dynamic programming matrix.

    Algorithm parameters:
        - gapCost (gap_cost) - cost of gap
        - matchPenalty (match_penalty) - penalty for alignment of two equal symbols
        - mismatchPenalty (mismatch_penalty) - penalty for alignment of two different symbols
    """
    name = "Operation-Weighted Edit Distance"
    parameters = {'gap_cost':         {'pos': 0,
                                       'type': int, 'min': 0, 'max': 100, 'default': 1, 'name': 'Gap cost'},
                  'match_penalty':    {'pos': 1,
                                       'type': int, 'min': 0, 'max': 100, 'default': 0, 'name': 'Match penalty'},
                  'mismatch_penalty': {'pos': 2,
                                       'type': int, 'min': 0, 'max': 100, 'default': 1, 'name': 'Mismatch penalty'}}

    def __init__(self):
        super(OWED, self).__init__()
        self.gap_cost = 1
        self.match_penalty = 0
        self.mismatch_penalty = 1

    def _get_initial_i(self, i):
        """
        Calculated as: i * gapCost
        """
        return i * self.gap_cost

    def _get_initial_j(self, j):
        """
        Calculated as: j * gapCost
        """
        return j * self.gap_cost

    def _calculate_substitution_score(self, c1, c2):
        """
        Returns matchPenalty on match (c1 = c2), mismatchPenalty on mismatch (c1 != c2).
        """
        if c1 == c2:
            return self.match_penalty
        else:
            return self.mismatch_penalty

    def _get_dp_matrix_value_at(self, i, j):
        """
        Calculated as: min(G(i - 1, j) + gapCost, G(i, j - 1) + gapCost, G(i - 1, j - 1) + d(i, j))
        """
        return min(self.dp_matrix[i - 1][j].value + self.gap_cost,
                   self.dp_matrix[i][j - 1].value + self.gap_cost,
                   self._calculate_diagonal_score(i, j))

    def _add_pointers_from_position_to_value(self, i, j, value):
        """
        Make pointers from G(i, j) to elements on left, right and on diagonal
        if they equal given value.
        """
        if (self.dp_matrix[i - 1][j].value + self.gap_cost) == value:
            self.dp_matrix[i][j].from_horizontal = True

        if (self.dp_matrix[i][j - 1].value + self.gap_cost) == value:
            self.dp_matrix[i][j].from_vertical = True

        if self._calculate_diagonal_score(i, j) == value:
            self.dp_matrix[i][j].from_diagonal = True

    def _construct_dp_matrix(self):
        """
        Iteratively fill the current matrix.
        """
        for i in range(1, self.dp_matrix.size_m):
            for j in range(1, self.dp_matrix.size_n):
                dp_matrix_max = self._get_dp_matrix_value_at(i, j)
                self.dp_matrix[i][j].value = dp_matrix_max
                self._add_pointers_from_position_to_value(i, j, dp_matrix_max)

    def run(self, sequence1, sequence2, gap_cost, match_penalty, mismatch_penalty):
        """
        Calculate alignments of sequence1 and sequence using given parameters.

        Resulting alignments are held in this instance's alignments attribute.
        """
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.gap_cost = gap_cost
        self.match_penalty = match_penalty
        self.mismatch_penalty = mismatch_penalty

        self.alignments = []
        self._initialize()
        self._construct_dp_matrix()
        self.backtrack(self.dp_matrix.size_m - 1, self.dp_matrix.size_n - 1)


class NeedlemanWunsch(DPAlgorithm):
    """
    Implementation of Needleman-Wunsch algorithm.

    Based on David Hoksza's bioinformatics lectures: http://siret.ms.mff.cuni.cz/members/hoksza/lectures

    Notation:
        - G matrix (dp_matrix) - dynamic programming matrix.
        - H matrix (h_matrix) - horizontal matrix for affine gap
        - V matrix (v_matrix) - vertical matrix for affine gap
        - s - scoring function (uses this instance's substitution_matrix)

    Algorithm parameters:
        - gapCost (gap_cost) - cost of gap (gap opening if affineGapCost is set)
        - affineGapCost (affine_gap_cost) - cost of gap extension (constant gap penalty used if not set)
    """
    name = "Needleman-Wunsch"
    parameters = {'gap_cost':        {'pos': 0,
                                      'type': int, 'min': 0, 'max': 100, 'default': 6, 'name': 'Gap cost'},
                  # use_affine parameter is required for UI only - toggles affine_gap_cost field
                  'use_affine':      {'pos': 1,
                                      'type': bool, 'name': 'Use affine gap', 'toggles': 'affine_gap_cost',
                                      'default': True},
                  'affine_gap_cost': {'pos': 2,
                                      'type': int, 'min': 0, 'max': 100, 'default': 3, 'name': 'Gap extension cost'},
                  }

    def __init__(self):
        super(NeedlemanWunsch, self).__init__()
        self.substitution_matrix = None

    def _get_initial_i(self, i):
        """
        Calculated as: - gapCost - (i - 1) * affineGapCost
        """
        return - self.gap_cost - (i - 1) * self.affine_gap_cost

    def _get_initial_j(self, j):
        """
        Calculated as: - gapCost - (j - 1) * affineGapCost
        """
        return - self.gap_cost - (j - 1) * self.affine_gap_cost

    def _initialize(self):
        """
        Matrices initialization - fills G, H and V with zeroes and calculates
        values for zeroth column and row of these matrices.
        """
        self.log.add_section('Initialization')
        self.dp_matrix = DPMatrix(len(self.sequence1) + 1, len(self.sequence2) + 1)
        self.h_matrix = [[0 for _ in range(self.dp_matrix.size_n)] for _ in range(self.dp_matrix.size_m)]
        self.v_matrix = [[0 for _ in range(self.dp_matrix.size_n)] for _ in range(self.dp_matrix.size_m)]

        self.log.append("Column 0 - " + self._get_initial_i.__doc__.lstrip())
        for i in range(1, self.dp_matrix.size_m):
            value = self._get_initial_i(i)
            self.dp_matrix[i][0].value = value
            self.dp_matrix[i][0].from_horizontal = True
            self.log.append('Set G(%s, 0) = H(%s, 0) = V(%s, 0) = %s' % (i, i, i, value))
            self.h_matrix[i][0] = value
            self.v_matrix[i][0] = value

        self.log.append("Row 0 - " + self._get_initial_j.__doc__.lstrip())
        for j in range(1, self.dp_matrix.size_n):
            value = self._get_initial_j(j)
            self.dp_matrix[0][j].value = value
            self.dp_matrix[0][j].from_vertical = True
            self.log.append('Set G(0, %s) = H(0, %s) = V(0, %s) = %s' % (j, j, j, value))
            self.h_matrix[0][j] = value
            self.v_matrix[0][j] = value

    def _calculate_substitution_score(self, c1, c2):
        """
        Calculate cost of substitution using substitution_matrix.
        """
        return self.substitution_matrix.get_score(c1, c2)

    def _get_dp_matrix_value_at(self, i, j):
        """
        Calculated as: max(H(i, j), V(i, j), G(i - 1, j - 1) + s(i, j))
        """
        return max(self.h_matrix[i][j], self.v_matrix[i][j], self._calculate_diagonal_score(i, j))

    def _get_h_matrix_value_at(self, i, j):
        """
        Calculated as: max(H(i - 1, j) - affineGapCost, G(i - 1, j) - gapCost)
        """
        return max(self.h_matrix[i - 1][j] - self.affine_gap_cost,
                   self.dp_matrix[i - 1][j].value - self.gap_cost)

    def _get_v_matrix_value_at(self, i, j):
        """
        Calculated as: max(V(i, j - 1) - affineGapCost, G(i, j - 1) - gapCost)
        """
        return max(self.v_matrix[i][j - 1] - self.affine_gap_cost,
                   self.dp_matrix[i][j - 1].value - self.gap_cost)

    def _add_pointers_from_position_to_value(self, i, j, value):
        """
        Make pointers from G(i, j) to elements on left, right and on diagonal
        if they equal given value in h_matrix, v_matrix or in diagonal element.
        """
        if self.h_matrix[i][j] == value:
            self.log.append("\t - equals H(i, j) - adding horizontal pointer")
            self.dp_matrix[i][j].from_horizontal = True

        if self.v_matrix[i][j] == value:
            self.log.append("\t - equals V(i, j) - adding vertical pointer")
            self.dp_matrix[i][j].from_vertical = True

        if self._calculate_diagonal_score(i, j) == value:
            self.log.append("\t - equals value on diagonal - adding diagonal pointer")
            self.dp_matrix[i][j].from_diagonal = True

    def _construct_dp_matrix(self):
        """
        Construct DP matrix - creates matrices G, H and V.
        """
        self.log.add_section('DP matrix construction')
        self.log.append("G(i, j) - %s" % self._get_dp_matrix_value_at.__doc__.lstrip())
        self.log.append("H(i, j) - %s" % self._get_h_matrix_value_at.__doc__.lstrip())
        self.log.append("V(i, j) - %s" % self._get_v_matrix_value_at.__doc__.lstrip())
        for i in range(1, self.dp_matrix.size_m):
            for j in range(1, self.dp_matrix.size_n):
                h_value = self._get_h_matrix_value_at(i, j)
                self.h_matrix[i][j] = h_value
                self.log.append("Set H(%s, %s) = %s" % (i, j, h_value))
                v_value = self._get_v_matrix_value_at(i, j)
                self.v_matrix[i][j] = v_value
                self.log.append("Set V(%s, %s) = %s" % (i, j, v_value))
                dp_matrix_max = self._get_dp_matrix_value_at(i, j)
                self.log.append("Set G(%s, %s) = %s" % (i, j, dp_matrix_max))
                self.dp_matrix[i][j].value = dp_matrix_max
                self._add_pointers_from_position_to_value(i, j, dp_matrix_max)

    def run(self, sequence1, sequence2, gap_cost, affine_gap_cost=None):
        """
        Calculate alignments of sequence1 and sequence2 with supplied gap_cost
        and optional affine_gap_cost if using affine gap.

        Resulting alignments are held in this instance's alignments attribute.
        Alignments start in bottom right of G matrix.
        """
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.gap_cost = gap_cost
        self.affine_gap_cost = affine_gap_cost or gap_cost

        self.alignments = []
        self._initialize()
        self._construct_dp_matrix()
        self.backtrack(self.dp_matrix.size_m - 1, self.dp_matrix.size_n - 1)


class SmithWaterman(NeedlemanWunsch):
    """
    Implementation of Smith-Waterman algorithm.

    Based on David Hoksza's bioinformatics lectures: http://siret.ms.mff.cuni.cz/members/hoksza/lectures

    Notation:
        - G matrix (dp_matrix) - dynamic programming matrix.
        - H matrix (h_matrix) - horizontal matrix for affine gap
        - V matrix (v_matrix) - vertical matrix for affine gap
        - s - scoring function (uses this instance's substitution_matrix)

    Algorithm parameters:
        - gapCost (gap_cost) - cost of gap (gap opening if affineGapCost is set)
        - affineGapCost (affine_gap_cost) - cost of gap extension (constant gap penalty used if not set)
    """
    name = "Smith-Waterman"

    def _get_initial_i(self, i):
        """
        Always zero.
        """
        return 0

    def _get_initial_j(self, j):
        """
        Always zero.
        """
        return 0

    def _get_dp_matrix_value_at(self, i, j):
        """
        Calculated as: max(H(i, j), V(i, j), G(i - 1, j - 1) + s(i, j), <b>0</b>)
        """
        return max(super(SmithWaterman, self)._get_dp_matrix_value_at(i, j),
                   0)

    def _is_backtracking_done(self, i, j):
        """
        Backtracking is done when current DP matrix element equals zero.
        """
        return self.dp_matrix[i][j].value == 0

    def run(self, sequence1, sequence2, gap_cost, affine_gap_cost=None):
        """
        Calculate alignments of sequence1 and sequence2 with supplied gap_cost
        and optional affine_gap_cost if using affine gap.

        Resulting alignments are held in this instance's alignments attribute.
        Alignments start in all elements that contain maximum value of G matrix.
        """
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.gap_cost = gap_cost
        self.affine_gap_cost = affine_gap_cost or gap_cost

        self.alignments = []
        self._initialize()
        self._construct_dp_matrix()
        for i, j in self.dp_matrix.get_max_positions():
            self.backtrack(i, j)
