import itertools
import os


class DPMatrixElement(object):
    """
    Element of DP matrix. Holds value and pointers to diagonal, horizontal
    and vertical elements.
    """
    def __init__(self):
        self.from_diagonal = False
        self.from_vertical = False
        self.from_horizontal = False
        self.value = 0


class DPMatrix(object):
    """
    Matrix of size m * n consisting of DPMatrixElements.
    """
    def __init__(self, size_m, size_n):
        self.size_m = size_m
        self.size_n = size_n
        self.matrix = []
        for i in range(0, size_m):
            self.matrix.append([])
            for j in range(0, size_n):
                self.matrix[i].append(DPMatrixElement())

    def __getitem__(self, index):
        return self.matrix[index]

    def __unicode__(self):
        result = u""
        for j in range(0, self.size_n):
            for i in range(0, self.size_m):
                result += "%4s" % self.matrix[i][j].value
            result += "\n"
        return result

    def __str__(self):
        return unicode(self).encode('utf-8')

    def _get_max_value(self):
        """
        Returns maximum value of element in this matrix as int.
        """
        max_element = max(itertools.chain.from_iterable(self.matrix), key=lambda x: x.value)
        return max_element.value

    def get_max_positions(self):
        """
        Returns list of tuples containing positions of maximum values.
        """
        result = []
        max_value = self._get_max_value()

        for i in range(0, self.size_m):
            for j in range(0, self.size_n):
                if self.matrix[i][j].value == max_value:
                    result.append((i, j))

        return result


class MatrixParsingException(Exception):
    """
    Raised when parsing of matrix file failed.
    """
    pass


class SubstitutionMatrix(object):
    """
    Substitution matrix used for scoring matches and mismatches.
    """
    def __init__(self):
        self._matrix = []
        self._indexes = {}
        self.name = ""

    def load_matrix(self, filename):
        """
        Load matrix from given filename and save the substitution scores
        into current instance.

        Matrix should be well-formed - order of columns and rows should be same,
        empty lines and lines starting with '#' are ignored - such matrices could be
        found at ftp://ftp.ncbi.nih.gov/blast/matrices/

        Raises MatrixParsingException when parsing fails.
        """
        line_no = 0
        matrix = []
        indexes = {}
        for line in open(filename):
            line = line.lstrip().rstrip()

            # skip blanks and comments
            if not line or line.startswith('#'):
                continue

            symbols = line.split()
            if line_no == 0:
                indexes = {symbol: i for i, symbol in enumerate(symbols)}
            else:
                # line should have symbol + len(indexes) elements
                if len(symbols) - 1 != len(indexes):
                    raise MatrixParsingException
                # order of rows should be same as order of columns
                try:
                    if line_no - 1 != indexes[symbols[0]]:
                        raise MatrixParsingException
                except KeyError:
                    raise MatrixParsingException
                # if the previous conditions passed, let's append the numbers to matrix
                matrix.append([int(i) for i in symbols[1:]])
            line_no += 1
        self._matrix = matrix
        self._indexes = indexes
        self.name = os.path.split(filename)[1]

    def _get_index(self, char):
        return self._indexes[char]

    def get_score(self, chr1, chr2):
        """
        Returns score of alignment of chr1 and chr2.
        """
        return self._matrix[self._get_index(chr1)][self._get_index(chr2)]
